#ifndef MSP_GL_BUFFER_H_
#define MSP_GL_BUFFER_H_

#include <vector>
#include "gl.h"
#include <msp/gl/extensions/arb_pixel_buffer_object.h>
#include <msp/gl/extensions/arb_vertex_buffer_object.h>
#include <msp/gl/extensions/arb_uniform_buffer_object.h>

namespace Msp {
namespace GL {

enum BufferType
{
	ARRAY_BUFFER         = GL_ARRAY_BUFFER,
	ELEMENT_ARRAY_BUFFER = GL_ELEMENT_ARRAY_BUFFER,
	PIXEL_PACK_BUFFER    = GL_PIXEL_PACK_BUFFER,
	PIXEL_UNPACK_BUFFER  = GL_PIXEL_UNPACK_BUFFER,
	UNIFORM_BUFFER       = GL_UNIFORM_BUFFER
};

enum BufferUsage
{
	STREAM_DRAW  = GL_STREAM_DRAW,
	STREAM_READ  = GL_STREAM_READ,
	STREAM_COPY  = GL_STREAM_COPY,
	STATIC_DRAW  = GL_STATIC_DRAW,
	STATIC_READ  = GL_STATIC_READ,
	STATIC_COPY  = GL_STATIC_COPY,
	DYNAMIC_DRAW = GL_DYNAMIC_DRAW,
	DYNAMIC_READ = GL_DYNAMIC_READ,
	DYNAMIC_COPY = GL_DYNAMIC_COPY
};

enum BufferAccess
{
	READ_ONLY = GL_READ_ONLY,
	WRITE_ONLY = GL_WRITE_ONLY,
	READ_WRITE = GL_READ_WRITE
};

class BufferRange;

/**
A buffer for storing data in GL memory.  Putting vertex and index data in
buffers can improve rendering performance.  The VertexArray, Mesh and
UniformBlock classes contain built-in support for buffers.
*/
class Buffer
{
	friend class BufferRange;

private:
	BufferType type;
	BufferUsage usage;
	unsigned id;
	unsigned size;

	static const Buffer *bound[5];

public:
	Buffer(BufferType);
	~Buffer();

private:
	static void require_buffer_type(BufferType);

public:
	/** Returns the OpenGL ID of the buffer.  For internal use only. */
	unsigned get_id() const { return id; }

	/** Returns the default binding type for the buffer. */
	BufferType get_type() const { return type; }

	/** Sets the usage hint of the buffer.  It will take effect the next time
	the buffer's contents are defined. */
	void set_usage(BufferUsage);

	/** Uploads data into the buffer, completely replacing any previous
	contents. */
	void data(unsigned, const void *);

	/** Overwrites part of the buffer data with new data.  The buffer size can
	not be changed with this call. */
	void sub_data(unsigned, unsigned, const void *);

	unsigned get_size() const { return size; }

	BufferRange *create_range(unsigned, unsigned);

	void *map(BufferAccess);
	bool unmap();

	/** Binds the buffer in its default slot. */
	void bind() const { bind_to(type); }

	/** Binds the buffer in an alternate slot. */
	void bind_to(BufferType) const;

	/** Unbinds the buffer from its default slot. */
	void unbind() const { unbind_from(type); }

	static const Buffer *current(BufferType);
	static void unbind_from(BufferType);
private:
	static const Buffer *&binding(BufferType);
	static bool set_current(BufferType, const Buffer *);
};


/**
A proxy for a subset of a buffer.  Can be bound for use with uniform blocks.
*/
class BufferRange
{
private:
	Buffer &buffer;
	unsigned offset;
	unsigned size;

	static std::vector<const BufferRange *> bound_uniform;

public:
	BufferRange(Buffer &, unsigned, unsigned);
	~BufferRange();

	void data(const void *);

	void bind_to(BufferType, unsigned);

	static const BufferRange *current(BufferType t, unsigned i) { return binding(t, i); }
	static void unbind_from(BufferType, unsigned);
private:
	static const BufferRange *&binding(BufferType, unsigned);
	static bool set_current(BufferType, unsigned, const BufferRange *);

public:
	static unsigned get_n_uniform_buffer_bindings();
	static unsigned get_uniform_buffer_alignment();
};

} // namespace GL
} // namespace Msp

#endif
