#include <msp/core/maputils.h>
#include "blend.h"
#include "camera.h"
#include "framebuffer.h"
#include "lighting.h"
#include "pipeline.h"
#include "postprocessor.h"
#include "renderbuffer.h"
#include "renderer.h"
#include "tests.h"
#include "texture2d.h"

using namespace std;

namespace Msp {
namespace GL {

Pipeline::Pipeline(unsigned w, unsigned h, bool d):
	camera(0),
	width(w),
	height(h),
	hdr(d),
	samples(0),
	target_ms(0),
	in_frame(false)
{
	target[0] = 0;
	target[1] = 0;
}

Pipeline::~Pipeline()
{
	delete target[0];
	delete target[1];
	delete target_ms;
}

void Pipeline::set_hdr(bool h)
{
	if(h==hdr)
		return;

	bool old_hdr= hdr;
	hdr = h;
	try
	{
		create_targets(2);
	}
	catch(...)
	{
		hdr = old_hdr;
		throw;
	}
}

void Pipeline::set_multisample(unsigned s)
{
	if(s==samples)
		return;

	unsigned old_samples = samples;
	samples = s;
	try
	{
		create_targets(1);
	}
	catch(...)
	{
		samples = old_samples;
		throw;
	}
}

void Pipeline::set_camera(const Camera *c)
{
	camera = c;
}

Pipeline::Pass &Pipeline::add_pass(const Tag &tag)
{
	passes.push_back(Pass(tag, 0));
	return passes.back();
}

void Pipeline::add_renderable(const Renderable &r)
{
	for(vector<Slot>::iterator i=renderables.begin(); i!=renderables.end(); ++i)
		if(i->renderable==&r)
		{
			i->passes.clear();
			return;
		}

	renderables.push_back(&r);
}

void Pipeline::add_renderable_for_pass(const Renderable &r, const Tag &tag)
{
	for(vector<Slot>::iterator i=renderables.begin(); i!=renderables.end(); ++i)
		if(i->renderable==&r)
		{
			i->passes.insert(tag);
			return;
		}

	renderables.push_back(&r);
	renderables.back().passes.insert(tag);
}

void Pipeline::remove_renderable(const Renderable &r)
{
	for(vector<Slot>::iterator i=renderables.begin(); i!=renderables.end(); ++i)
		if(i->renderable==&r)
		{
			renderables.erase(i);
			return;
		}
}

Pipeline::Pass &Pipeline::add_pass(const Tag &tag, const Renderable &r)
{
	passes.push_back(Pass(tag, &r));
	return passes.back();
}

void Pipeline::add_postprocessor(PostProcessor &pp)
{
	postproc.push_back(&pp);
	try
	{
		create_targets(0);
	}
	catch(...)
	{
		postproc.pop_back();
		throw;
	}
}

void Pipeline::setup_frame() const
{
	in_frame = true;
	for(PassList::const_iterator i=passes.begin(); i!=passes.end(); ++i)
		if(const Renderable *renderable = i->get_renderable())
			renderable->setup_frame();
	for(vector<Slot>::const_iterator i=renderables.begin(); i!=renderables.end(); ++i)
		i->renderable->setup_frame();
}

void Pipeline::finish_frame() const
{
	in_frame = false;
	for(PassList::const_iterator i=passes.begin(); i!=passes.end(); ++i)
		if(const Renderable *renderable = i->get_renderable())
			renderable->finish_frame();
	for(vector<Slot>::const_iterator i=renderables.begin(); i!=renderables.end(); ++i)
		i->renderable->finish_frame();
}

void Pipeline::render(const Tag &tag) const
{
	if(tag.id)
		return;

	Renderer renderer(camera);
	render(renderer, tag);
}

void Pipeline::render(Renderer &renderer, const Tag &tag) const
{
	if(tag.id)
		return;

	bool was_in_frame = in_frame;
	if(!in_frame)
		setup_frame();

	const Framebuffer *out_fbo = Framebuffer::current();
	// These is a no-ops but will ensure the related state gets restored
	BindRestore restore_fbo(out_fbo);
	BindRestore restore_depth_test(DepthTest::current());
	BindRestore restore_blend(Blend::current());

	if(target[0])
	{
		Framebuffer &fbo = (samples ? target_ms : target[0])->get_framebuffer();
		fbo.bind();
		fbo.clear(COLOR_BUFFER_BIT|DEPTH_BUFFER_BIT);
	}


	for(PassList::const_iterator i=passes.begin(); i!=passes.end(); ++i)
	{
		if(const DepthTest *dt = i->get_depth_test())
			dt->bind();
		else
			DepthTest::unbind();

		if(const Blend *b = i->get_blend())
			b->bind();
		else
			Blend::unbind();

		renderer.set_lighting(i->get_lighting());
		renderer.set_clipping(i->get_clipping());

		if(const Renderable *renderable = i->get_renderable())
			renderer.render(*renderable, i->get_tag());

		for(vector<Slot>::const_iterator j=renderables.begin(); j!=renderables.end(); ++j)
			if(j->passes.empty() || j->passes.count(i->get_tag()))
				renderer.render(*j->renderable, i->get_tag());
	}

	if(target[0])
	{
		DepthTest::unbind();
		Blend::unbind();

		if(samples)
			target[0]->blit_from(*target_ms);

		for(unsigned i=0; i<postproc.size(); ++i)
		{
			unsigned j = i%2;
			if(i+1<postproc.size())
				target[1-j]->get_framebuffer().bind();
			else
				out_fbo->bind();
			const Texture2D &color = target[j]->get_target_texture(RENDER_COLOR);
			const Texture2D &depth = target[j]->get_target_texture(RENDER_DEPTH);
			postproc[i]->render(renderer, color, depth);
		}
	}

	if(!was_in_frame)
		finish_frame();
}

void Pipeline::create_targets(unsigned recreate)
{
	if(recreate>=2)
	{
		delete target[0];
		delete target[1];
		target[0] = 0;
		target[1] = 0;
	}
	if(recreate>=1)
	{
		delete target_ms;
		target_ms = 0;
	}

	PixelFormat color_pf = (hdr ? RGB16F : RGB);
	RenderTargetFormat fmt = (RENDER_COLOR,color_pf, RENDER_DEPTH);
	if(!postproc.empty() || samples)
	{
		if(!target[0])
			target[0] = new RenderTarget(width, height, fmt);
		if(!target[1] && postproc.size()>1)
			target[1] = new RenderTarget(width, height, fmt);
	}

	if(!target_ms && samples)
		target_ms = new RenderTarget(width, height, samples, fmt);
}


Pipeline::Pass::Pass(const Tag &t, const Renderable *r):
	tag(t),
	lighting(0),
	depth_test(0),
	blend(0),
	clipping(0),
	renderable(r)
{ }

void Pipeline::Pass::set_lighting(const Lighting *l)
{
	lighting = l;
}

void Pipeline::Pass::set_depth_test(const DepthTest *d)
{
	depth_test = d;
}

void Pipeline::Pass::set_blend(const Blend *b)
{
	blend = b;
}

void Pipeline::Pass::set_clipping(const Clipping *c)
{
	clipping =c;
}


Pipeline::Slot::Slot(const Renderable *r):
	renderable(r)
{ }

} // namespace GL
} // namespace Msp
