#ifndef MSP_GL_PROGRAMCOMPILER_H_
#define MSP_GL_PROGRAMCOMPILER_H_

#include <set>
#include "programparser.h"
#include "programsyntax.h"

namespace Msp {
namespace GL {

class Program;
class Resources;

class ProgramCompiler
{
private:
	struct Visitor: ProgramSyntax::TraversingVisitor
	{
		typedef void ResultType;

		ProgramSyntax::Stage *stage;

		Visitor();

		virtual void apply(ProgramSyntax::Stage &);
		void get_result() const { }
	};

	struct Formatter: Visitor
	{
		typedef std::string ResultType;

		std::string formatted;
		unsigned indent;
		bool parameter_list;
		unsigned else_if;
		std::string block_interface;

		Formatter();

		virtual void apply(ProgramSyntax::Stage &);
		const std::string &get_result() const { return formatted; }
		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::Literal &);
		virtual void visit(ProgramSyntax::ParenthesizedExpression &);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::MemberAccess &);
		virtual void visit(ProgramSyntax::UnaryExpression &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::Assignment &);
		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::ExpressionStatement &);
		virtual void visit(ProgramSyntax::Import &);
		virtual void visit(ProgramSyntax::Layout &);
		virtual void visit(ProgramSyntax::InterfaceLayout &);
		virtual void visit(ProgramSyntax::StructDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
		virtual void visit(ProgramSyntax::Conditional &);
		virtual void visit(ProgramSyntax::Iteration &);
		virtual void visit(ProgramSyntax::Return &);
	};

	template<typename T>
	struct NodeGatherer: Visitor
	{
		typedef std::list<T *> ResultType;

		std::list<T *> nodes;

		const ResultType &get_result() const { return nodes; }
		virtual void visit(T &n) { nodes.push_back(&n); }
	};

	struct DeclarationCombiner: Visitor
	{
		bool toplevel;
		std::map<std::string, std::vector<ProgramSyntax::FunctionDeclaration *> > functions;
		std::map<std::string, ProgramSyntax::VariableDeclaration *> variables;
		bool remove_node;

		DeclarationCombiner();

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
	};

	struct VariableResolver: Visitor
	{
		std::vector<ProgramSyntax::Block *> blocks;
		ProgramSyntax::StructDeclaration *type;
		bool anonymous;
		std::string block_interface;
		bool record_target;
		ProgramSyntax::VariableDeclaration *assignment_target;
		bool self_referencing;

		VariableResolver();

		virtual void apply(ProgramSyntax::Stage &);
		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::MemberAccess &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::Assignment &);
		virtual void visit(ProgramSyntax::StructDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
	};

	struct FunctionResolver: Visitor
	{
		std::map<std::string, std::vector<ProgramSyntax::FunctionDeclaration *> > functions;

		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
	};

	struct BlockModifier: Visitor
	{
		bool remove_node;
		std::list<RefPtr<ProgramSyntax::Node> > insert_nodes;

		BlockModifier();

		void flatten_block(ProgramSyntax::Block &);
		void apply_and_increment(ProgramSyntax::Block &, std::list<RefPtr<ProgramSyntax::Node> >::iterator &);
		virtual void visit(ProgramSyntax::Block &);
	};

	struct InterfaceGenerator: BlockModifier
	{
		std::string in_prefix;
		std::string out_prefix;
		unsigned scope_level;
		std::map<std::string, RefPtr<ProgramSyntax::VariableDeclaration> > iface_declarations;

		InterfaceGenerator();

		static std::string get_out_prefix(ProgramSyntax::StageType);
		virtual void apply(ProgramSyntax::Stage &);
		virtual void visit(ProgramSyntax::Block &);
		std::string change_prefix(const std::string &, const std::string &) const;
		bool generate_interface(ProgramSyntax::VariableDeclaration &, const std::string &, const std::string &);
		void insert_assignment(const std::string &, ProgramSyntax::Expression *);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::Passthrough &);
	};

	struct VariableRenamer: Visitor
	{
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
	};

	struct DeclarationReorderer: Visitor
	{
		enum DeclarationKind
		{
			NO_DECLARATION,
			LAYOUT,
			STRUCT,
			VARIABLE,
			FUNCTION
		};

		DeclarationKind kind;

		DeclarationReorderer();

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::InterfaceLayout &) { kind = LAYOUT; }
		virtual void visit(ProgramSyntax::StructDeclaration &) { kind = STRUCT; }
		virtual void visit(ProgramSyntax::VariableDeclaration &) { kind = VARIABLE; }
		virtual void visit(ProgramSyntax::InterfaceBlock &) { kind = VARIABLE; }
		virtual void visit(ProgramSyntax::FunctionDeclaration &) { kind = FUNCTION; }
	};

	struct InlineableFunctionLocator: Visitor
	{
		typedef std::set<ProgramSyntax::FunctionDeclaration *> ResultType;

		std::map<ProgramSyntax::FunctionDeclaration *, unsigned> refcounts;
		std::set<ProgramSyntax::FunctionDeclaration *> inlineable;
		ProgramSyntax::FunctionDeclaration *in_function;

		InlineableFunctionLocator();

		const ResultType &get_result() const { return inlineable; }
		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
	};

	struct FunctionInliner: Visitor
	{
		std::set<ProgramSyntax::FunctionDeclaration *> inlineable;
		unsigned extract_result;
		RefPtr<ProgramSyntax::Expression> inline_result;

		FunctionInliner();
		FunctionInliner(const std::set<ProgramSyntax::FunctionDeclaration *> &);

		void visit_and_inline(RefPtr<ProgramSyntax::Expression> &);
		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::UnaryExpression &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::Return &);
	};

	struct ExpressionEvaluator: ProgramSyntax::NodeVisitor
	{
		typedef std::map<ProgramSyntax::VariableDeclaration *, ProgramSyntax::Expression *> ValueMap;

		const ValueMap *variable_values;
		float result;
		bool result_valid;

		ExpressionEvaluator();
		ExpressionEvaluator(const ValueMap &);

		virtual void visit(ProgramSyntax::Literal &);
		virtual void visit(ProgramSyntax::ParenthesizedExpression &);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::UnaryExpression &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
	};

	struct ConstantConditionEliminator: BlockModifier
	{
		unsigned scope_level;
		ExpressionEvaluator::ValueMap variable_values;

		ConstantConditionEliminator();

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::Assignment &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::Conditional &);
		virtual void visit(ProgramSyntax::Iteration &);
	};

	struct UnusedVariableLocator: Visitor
	{
		struct AssignmentList
		{
			std::vector<ProgramSyntax::Node *> nodes;
			bool conditional;
			bool self_referencing;
		};

		typedef std::set<ProgramSyntax::Node *> ResultType;
		typedef std::map<ProgramSyntax::VariableDeclaration *, AssignmentList> BlockAssignmentMap;

		std::set<ProgramSyntax::Node *> unused_nodes;
		std::map<ProgramSyntax::VariableDeclaration *, ProgramSyntax::Node *> aggregates;
		ProgramSyntax::Node *aggregate;
		std::vector<BlockAssignmentMap> assignments;
		ProgramSyntax::Assignment *assignment;
		bool assignment_target;

		UnusedVariableLocator();

		virtual void apply(ProgramSyntax::Stage &);
		const ResultType &get_result() const { return unused_nodes; }
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::MemberAccess &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::Assignment &);
		void record_assignment(ProgramSyntax::VariableDeclaration &, ProgramSyntax::Node &, bool);
		virtual void visit(ProgramSyntax::ExpressionStatement &);
		virtual void visit(ProgramSyntax::StructDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
		void merge_down_assignments();
		virtual void visit(ProgramSyntax::Conditional &);
		virtual void visit(ProgramSyntax::Iteration &);
	};

	struct UnusedFunctionLocator: Visitor
	{
		typedef std::set<ProgramSyntax::Node *> ResultType;

		std::set<ProgramSyntax::Node *> unused_nodes;
		std::set<ProgramSyntax::FunctionDeclaration *> used_definitions;

		const ResultType &get_result() const { return unused_nodes; }
		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
	};

	struct NodeRemover: Visitor
	{
		std::set<ProgramSyntax::Node *> to_remove;

		NodeRemover() { }
		NodeRemover(const std::set<ProgramSyntax::Node *> &);

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
	};

	struct LegacyConverter: BlockModifier
	{
		Version target_version;
		std::string type;
		std::string frag_out_name;

		LegacyConverter();
		LegacyConverter(const Version &);

		bool check_version(const Version &);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
	};

	Resources *resources;
	ProgramParser parser;
	ProgramSyntax::Module *module;

public:
	ProgramCompiler();

	void compile(const std::string &);
	void compile(IO::Base &, Resources * = 0);
	void add_shaders(Program &);

private:
	static ProgramSyntax::Module *create_builtins_module();
	static ProgramSyntax::Module &get_builtins_module();
	static ProgramSyntax::Stage *get_builtins(ProgramSyntax::StageType);
	void process();
	void import(const std::string &);
	void generate(ProgramSyntax::Stage &);
	bool optimize(ProgramSyntax::Stage &);
	static void inject_block(ProgramSyntax::Block &, const ProgramSyntax::Block &);
	template<typename T>
	static typename T::ResultType apply(ProgramSyntax::Stage &);
	template<typename T, typename A>
	static typename T::ResultType apply(ProgramSyntax::Stage &, const A &);
};

} // namespace GL
} // namespace Msp

#endif
